/*
// ============================================================================
//
// = LIBRARY
//     OTC
// 
// = FILENAME
//     dispatch/otcsvbroker.cc
//
// = AUTHOR(S)
//     Graham Dumpleton
// 
// = COPYRIGHT
//     Copyright 1999-2001 DUMPLETON SOFTWARE CONSULTING PTY LIMITED
//
// ============================================================================
*/

#ifdef __GNUG__
#pragma implementation "OTC/dispatch/svbroker.hh"
#endif

#include <OTC/dispatch/svbroker.hh>
#include <OTC/dispatch/report.hh>
#include <OTC/dispatch/request.hh>
#include <OTC/dispatch/response.hh>
#include <OTC/dispatch/failure.hh>
#include <OTC/dispatch/envelope.hh>
#include <OTC/dispatch/servannc.hh>
#include <OTC/dispatch/pbntfctn.hh>
#include <OTC/dispatch/sbscrptn.hh>
#include <OTC/text/symbol.hh>
#include <OTC/debug/tracer.hh>
#include <OTC/debug/tracetag.hh>
#include <OTC/debug/logger.hh>
#include <OTC/debug/logstrm.hh>
#include <OTC/thread/mxreaper.hh>

#include <OTC/stdlib/strstream.h>

#include <time.h>

/* ------------------------------------------------------------------------- */
static OTC_NRMutex MODULE_TRACETAG_MUTEX;
static OTC_TraceTag& MODULE_TRACETAG()
{
  static OTC_TraceTag* TAG = 0;
  MODULE_TRACETAG_MUTEX.lock();
  if (TAG == 0)
    TAG = new OTC_TraceTag("OTC_SVBroker");
  MODULE_TRACETAG_MUTEX.unlock();
  return *TAG;
}

#define MODULE MODULE_TRACETAG()

/* ------------------------------------------------------------------------- */
OTC_NRMutex OTC_SVBroker::mutex_;
OTC_SVRegistry* OTC_SVBroker::gRegistry_ = 0;

/* ------------------------------------------------------------------------- */
class OTC_ReportCache : public OTC_MPObject
{
  public:

                        ~OTC_ReportCache() {}

                        OTC_ReportCache(
                         OTC_String const& theContent,
                         char const* theContentType,
                         time_t thePostingTime,
                         time_t theExpirationTime
                        )
                          : content_(theContent),
                            contentType_(theContentType),
                            postingTime_(thePostingTime),
                            expirationTime_(theExpirationTime) {}

    OTC_String const&	content() const
				{ return content_; }

    OTC_String const&	contentType() const
				{ return contentType_; }

    time_t              postingTime() const
                                { return postingTime_; }

    time_t              expirationTime() const
                                { return expirationTime_; }

  private:

    OTC_String		content_;

    OTC_String		contentType_;

    time_t              postingTime_;

    time_t              expirationTime_;
};

/* ------------------------------------------------------------------------- */
OTC_SVRegistry* OTC_SVBroker::registry()
{
  mutex_.lock();

  OTC_MutexReaper<OTC_NRMutex> xxxMutex;
  xxxMutex.grab(mutex_);

  if (gRegistry_ == 0)
  {
    gRegistry_ = new OTC_SVRegistry;
    OTCLIB_ASSERT_M(gRegistry_ != 0);
  }

  xxxMutex.release();

  mutex_.unlock();

  return gRegistry_;
}

/* ------------------------------------------------------------------------- */
OTC_SVBroker::~OTC_SVBroker()
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_SVBroker::~OTC_SVBroker()");

  // Remove all cached reports.

  expireReports(0);

  // Remove all registry subscriptions.

  registry()->unsubscribeAgent(id());

  // Remove service binding.

  registry()->removeServiceBinding(serviceBinding_->serviceAddress());
}

/* ------------------------------------------------------------------------- */
OTC_SVBroker::OTC_SVBroker(
 OTC_EVAgent* theAgent,
 char const* theServiceName,
 char const* theServiceAudience
)
  : agent_(theAgent), serviceBinding_(0)
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_SVBroker::OTC_SVBroker()");

  OTCLIB_ENSURE_FN((theAgent != 0),
   "OTC_SVBroker::OTC_SVBroker()",
   "invalid service agent");

  // Create anonymous receiver.

  inBox_.bind(this);

  // Create binding.

  serviceBinding_ = registry()->addServiceBinding(theServiceName,
   theServiceAudience,inBox_.name(),"","");

  OTCLIB_TRACER(MODULE) << "serviceName = "
   << serviceBinding_->serviceName() << endl;
  OTCLIB_TRACER(MODULE) << "serviceAddress = "
   << serviceBinding_->serviceAddress() << endl;
  OTCLIB_TRACER(MODULE) << "serviceAudience = "
   << serviceBinding_->serviceAudience() << endl;
}

/* ------------------------------------------------------------------------- */
void OTC_SVBroker::handle(OTC_Event* theEvent)
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_SVBroker::handle()");

  if (theEvent == 0)
    return;

  OTCLIB_TRACER(MODULE) << "event = " << *theEvent << endl;

  if (theEvent->type() == OTCEV_ServiceAnnouncement::typeId())
  {
    OTCEV_ServiceAnnouncement* theAnnouncement;
    theAnnouncement = (OTCEV_ServiceAnnouncement*) theEvent;

    if (theAnnouncement->subscriptionType() == OTCLIB_SERVICE_NAME)
    {
      OTC_ServiceBinding* theBinding;
      theBinding = theAnnouncement->serviceBinding();

      OTC_String theKey;

      theKey += theBinding->serviceName();

      mutex_.lock();

      OTC_MutexReaper<OTC_NRMutex> xxxMutex;
      xxxMutex.grab(mutex_);

      if (theAnnouncement->status() == OTCLIB_SERVICE_AVAILABLE)
      {
	if (!publishers_.contains(theKey,theBinding))
	{
	  publishers_.add(theKey,theBinding);

	  OTC_Iterator<OTC_String> theSubjectPatterns;
	  theSubjectPatterns = subscriptions_.items(theKey);

	  while (theSubjectPatterns.isValid())
	  {
	    // Send subscription request.

	    OTC_String theHeader;
	    theHeader += "SubjectPattern: ";
	    theHeader += theSubjectPatterns.item();
	    theHeader += EOL;

	    OTC_Event* theMessage;
	    theMessage = new OTCEV_Message("");
	    OTCLIB_ASSERT_M(theMessage != 0);

	    OTCEV_Envelope* theEnvelope;
	    theEnvelope = new OTCEV_Envelope(theMessage,
	     theBinding->serviceAddress(),
	     serviceBinding_->serviceAddress(),
	     "$S::SUBSCRIPTION-REQUESTED",theHeader);
	    OTCLIB_ASSERT_M(theEnvelope != 0);

	    theEnvelope->queue(OTCLIB_STANDARD_JOB);

	    theSubjectPatterns.next();
	  }

	  OTCEV_PublisherNotification* theNotification;

	  theNotification = new OTCEV_PublisherNotification(
	   theBinding,serviceBinding_,theAnnouncement->status());
	  OTCLIB_ASSERT_M(theNotification != 0);

	  theNotification->queue(agent_->id(),OTCLIB_PRIORITY_JOB);
	}
      }
      else
      {
	if (publishers_.containsItem(theBinding))
	{
	  publishers_.removeItem(theBinding);

	  OTCEV_PublisherNotification* theNotification;

	  theNotification = new OTCEV_PublisherNotification(
	   theBinding,serviceBinding_,theAnnouncement->status());
	  OTCLIB_ASSERT_M(theNotification != 0);

	  theNotification->queue(agent_->id(),OTCLIB_PRIORITY_JOB);
	}
      }

      xxxMutex.release();

      mutex_.unlock();
    }
    else if (theAnnouncement->subscriptionType() == OTCLIB_SERVICE_ADDRESS)
    {
      OTC_ServiceBinding* theBinding;
      theBinding = theAnnouncement->serviceBinding();

      mutex_.lock();

      OTC_MutexReaper<OTC_NRMutex> xxxMutex;
      xxxMutex.grab(mutex_);

      if (theAnnouncement->status() == OTCLIB_SERVICE_WITHDRAWN)
      {
	if (subscribers_.containsItem(theBinding))
	{
	  registry()->unsubscribeServiceAddress(
	   id(),theBinding->serviceAddress());

	  OTC_Iterator<OTC_String> theSubjectPatterns;
	  theSubjectPatterns = subscribers_.keys(theBinding);

	  while (theSubjectPatterns.isValid())
	  {
	    OTCEV_Subscription* theSubscription;

	    theSubscription = new OTCEV_Subscription(
	     serviceBinding_,theAnnouncement->serviceBinding(),
	     OTCLIB_SUBSCRIPTION_WITHDRAWN,theSubjectPatterns.item());
	    OTCLIB_ASSERT_M(theSubscription != 0);

	    theSubscription->queue(agent_->id(),OTCLIB_PRIORITY_JOB);

	    theSubjectPatterns.next();
	  }

	  subscribers_.removeItem(theBinding);
	}
      }

      xxxMutex.release();

      mutex_.unlock();
    }
  }
  else if (theEvent->type() == OTCEV_Envelope::typeId())
  {
    OTCEV_Envelope* theEnvelope;
    theEnvelope = (OTCEV_Envelope*)theEvent;

    if (theEnvelope->subject() == "$S::REPORT")
    {
      OTC_Iterator<OTC_ServiceBinding*> thePublisherHandle;
      thePublisherHandle = registry()->lookupServiceAddress(
       theEnvelope->from());

      if (thePublisherHandle.isValid())
      {
	OTCEV_Report* thePublication;

	thePublication = new OTCEV_Report(theEnvelope,
	 thePublisherHandle.item(),serviceBinding_);
	OTCLIB_ASSERT_M(thePublication);

	thePublication->queue(agent_->id(),OTCLIB_PRIORITY_JOB);

	OTCLIB_TRACER(MODULE) << "serviceName = " <<
	 thePublisherHandle.item()->serviceName() << endl;
	OTCLIB_TRACER(MODULE) << "subjectName = " <<
	 thePublication->subjectName() << endl;
      }
    }
    else if (theEnvelope->subject() == "$S::REQUEST")
    {
      if (serviceBinding_->serviceAddress() == theEnvelope->to())
      {
	OTCEV_Request* theRequest;

	theRequest = new OTCEV_Request(theEnvelope,serviceBinding_);
	OTCLIB_ASSERT_M(theRequest);

	if (!builtinRequest_(theRequest))
	  theRequest->queue(agent_->id(),OTCLIB_PRIORITY_JOB);
      }
    }
    else if (theEnvelope->subject() == "$S::RESPONSE")
    {
      if (serviceBinding_->serviceAddress() == theEnvelope->to())
      {
	OTCEV_Response* theResponse;

	theResponse = new OTCEV_Response(theEnvelope);
	OTCLIB_ASSERT_M(theResponse);

	theResponse->queue(agent_->id(),OTCLIB_PRIORITY_JOB);
      }
    }
    else if (theEnvelope->subject() == "$S::FAILURE")
    {
      if (serviceBinding_->serviceAddress() == theEnvelope->to())
      {
	OTCEV_Failure* theFailure;

	theFailure = new OTCEV_Failure(theEnvelope);
	OTCLIB_ASSERT_M(theFailure);

	theFailure->queue(agent_->id(),OTCLIB_PRIORITY_JOB);
      }
    }
    else if (theEnvelope->subject() == "$S::SUBSCRIPTION-REQUESTED" ||
     theEnvelope->subject() == "$S::SUBSCRIPTION-WITHDRAWN")
    {
      mutex_.lock();

      OTC_MutexReaper<OTC_NRMutex> xxxMutex;
      xxxMutex.grab(mutex_);

      if (serviceBinding_->serviceAddress() == theEnvelope->to())
      {
	OTC_String theHeader;
	theHeader = theEnvelope->header();

	istrstream ins1((char*)theHeader.string(),theHeader.length());

	OTC_String tmpLine;
	OTC_String theSubjectPattern;

	while (ins1.good())
	{
	  tmpLine.readToDelim(ins1,EOL);
	  if (!ins1.fail())
	  {
	    if (tmpLine.compare(0,"SubjectPattern: ",16))
	      theSubjectPattern = tmpLine.after(15);
	  }
	}

	OTC_Iterator<OTC_ServiceBinding*> theSubscriberHandle;
	theSubscriberHandle = registry()->lookupServiceAddress(
	 theEnvelope->from());

        if (!theSubjectPattern.isUndefined() && theSubscriberHandle.isValid())
	{
	  if (theEnvelope->subject() == "$S::SUBSCRIPTION-REQUESTED")
	  {
	    OTCEV_Subscription* theSubscription;

	    theSubscription = new OTCEV_Subscription(
	     serviceBinding_,theSubscriberHandle.item(),
	     OTCLIB_SUBSCRIPTION_REQUESTED,theSubjectPattern);
	    OTCLIB_ASSERT_M(theSubscription != 0);

	    theSubscription->queue(agent_->id(),OTCLIB_PRIORITY_JOB);

	    if (!subscribers_.containsItem(theSubscriberHandle.item()))
	    {
	      registry()->subscribeServiceAddress(id(),
	       theSubscriberHandle.item()->serviceAddress());
	    }

	    if (!subscribers_.contains(theSubjectPattern,
	     theSubscriberHandle.item()))
	    {
	      subscribers_.add(theSubjectPattern,theSubscriberHandle.item());

	      // Send MATCHING cached events which haven't expired.

	      OTC_HIndexEntry<OTC_String>* theEntry;
	      theEntry = subjects_.firstEntry();

	      time_t theCurrentTime;
	      theCurrentTime = time(0);

	      while (theEntry != 0)
	      {
		if (matchSubject(theSubjectPattern,theEntry->key()))
		{
		  OTC_ReportCache* theCacheEntry;
		  theCacheEntry = (OTC_ReportCache*)theEntry->item();

		  if (theCacheEntry->expirationTime() != -1 &&
		   theCurrentTime > theCacheEntry->expirationTime())
		  {
		    delete theCacheEntry;

		    OTC_HIndexEntry<OTC_String>* theOldEntry;
		    theOldEntry = theEntry;

		    theEntry = theEntry->nextEntry();

		    theOldEntry->removeEntry();
		  }
		  else
		  {
		    // Publish event.

		    OTC_String theHeader;
		    theHeader += "SubjectName: ";
		    theHeader += theEntry->key();
		    theHeader += EOL;

		    OTCEV_Message* theMessage;
		    theMessage = new OTCEV_Message(
		     theCacheEntry->content(),theCacheEntry->contentType());
		    OTCLIB_ASSERT_M(theMessage != 0);

		    OTCEV_Envelope* theEnvelope;
		    theEnvelope = new OTCEV_Envelope(theMessage,
		     theSubscriberHandle.item()->serviceAddress(),
		     serviceBinding_->serviceAddress(),
		     "$S::REPORT",theHeader);
		    OTCLIB_ASSERT_M(theEnvelope != 0);

		    theEnvelope->queue(OTCLIB_STANDARD_JOB);

		    theEntry = theEntry->nextEntry();
		  }
		}
		else
		{
		  theEntry = theEntry->nextEntry();
		}
	      }
	    }
	  }
	  else
	  {
	    if (subscribers_.contains(theSubjectPattern,
	     theSubscriberHandle.item()))
	    {
	      OTCEV_Subscription* theSubscription;

	      theSubscription = new OTCEV_Subscription(
	       serviceBinding_,theSubscriberHandle.item(),
	       OTCLIB_SUBSCRIPTION_WITHDRAWN,theSubjectPattern);
	      OTCLIB_ASSERT_M(theSubscription != 0);

	      theSubscription->queue(agent_->id(),OTCLIB_PRIORITY_JOB);

	      subscribers_.remove(theSubjectPattern,
	       theSubscriberHandle.item());

	      if (!subscribers_.containsItem(theSubscriberHandle.item()))
	      {
		registry()->unsubscribeServiceAddress(id(),
		 theSubscriberHandle.item()->serviceAddress());
	      }
	    }
	  }
	}
      }

      xxxMutex.release();

      mutex_.unlock();
    }
  }

  theEvent->destroy();
}

/* ------------------------------------------------------------------------- */
void OTC_SVBroker::joinGroup(char const* theServiceGroup)
{
  registry()->addToServiceGroup(theServiceGroup,
   serviceBinding_->serviceAddress());
}

/* ------------------------------------------------------------------------- */
void OTC_SVBroker::leaveGroup(char const* theServiceGroup)
{
  registry()->removeFromServiceGroup(theServiceGroup,
   serviceBinding_->serviceAddress());
}

/* ------------------------------------------------------------------------- */
OTC_String OTC_SVBroker::sendRequest(
 OTC_ServiceBinding* theServiceAgent,
 char const* theMethodName,
 OTC_ROPayload const& theContent
)
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_SVBroker::sendRequest()");

  if (theMethodName == 0)
    theMethodName = "";

  if (theServiceAgent == 0)
  {
    OTC_LogStream theLogger;
    theLogger.setLevel(OTCLIB_LOG_WARNING);

    theLogger << "Precondition failed" << EOL;
    theLogger << "Class: OTC_SVBroker" << EOL;
    theLogger << "Method: sendRequest()" << EOL;
    theLogger << "Description: service is a null pointer" << EOL;
    theLogger << "Result: the request will be ignored" << EOL;
    theLogger << "ReceiverName: " << inBox_.name() << EOL;
    theLogger << "MethodName: " << theMethodName << EOL;
    theLogger << flush;

    return OTC_String::undefinedString();
  }

  // Send the request.

  OTC_String tmpContent;
  char const* tmpContentType;

  tmpContent = theContent.encode();
  tmpContentType = "text/xml";

  OTC_String theHeader;

  theHeader += "MethodName: ";
  theHeader += theMethodName;
  theHeader += EOL;

  OTCEV_Message* theMessage;
  theMessage = new OTCEV_Message(tmpContent,tmpContentType);
  OTCLIB_ASSERT_M(theMessage != 0);

  OTCEV_Envelope* theEnvelope;
  theEnvelope = new OTCEV_Envelope(
   theMessage,theServiceAgent->serviceAddress(),
   serviceBinding_->serviceAddress(),"$S::REQUEST",theHeader);
  OTCLIB_ASSERT_M(theEnvelope != 0);

  OTC_String theConversationId;
  theConversationId = theEnvelope->conversationId();

  OTCLIB_TRACER(MODULE) << "sendRequest =" << *theEnvelope << endl;
  OTCLIB_TRACER(MODULE) << "conversationId =" << theConversationId << endl;

  theEnvelope->queue(OTCLIB_STANDARD_JOB);

  return theConversationId;
}

/* ------------------------------------------------------------------------- */
OTC_String OTC_SVBroker::sendRequest(
 char const* theServiceAddress,
 char const* theMethodName,
 OTC_ROPayload const& theContent
)
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_SVBroker::sendRequest()");

  if (theMethodName == 0)
    theMethodName = "";

  if (theServiceAddress == 0 || *theServiceAddress == EOS)
  {
    OTC_LogStream theLogger;
    theLogger.setLevel(OTCLIB_LOG_WARNING);

    theLogger << "Precondition failed" << EOL;
    theLogger << "Class: OTC_SVBroker" << EOL;
    theLogger << "Method: sendRequest()" << EOL;
    theLogger << "Description: service address is empty" << EOL;
    theLogger << "Result: the request will be ignored" << EOL;
    theLogger << "ReceiverName: " << inBox_.name() << EOL;
    theLogger << "MethodName: " << theMethodName << EOL;
    theLogger << flush;

    return OTC_String::undefinedString();
  }

  // Send the request.

  OTC_String tmpContent;
  char const* tmpContentType;

  tmpContent = theContent.encode();
  tmpContentType = "text/xml";

  OTC_String theHeader;

  theHeader += "MethodName: ";
  theHeader += theMethodName;
  theHeader += EOL;

  OTCEV_Message* theMessage;
  theMessage = new OTCEV_Message(tmpContent,tmpContentType);
  OTCLIB_ASSERT_M(theMessage != 0);

  OTCEV_Envelope* theEnvelope;
  theEnvelope = new OTCEV_Envelope(
   theMessage,theServiceAddress,
   serviceBinding_->serviceAddress(),"$S::REQUEST",theHeader);
  OTCLIB_ASSERT_M(theEnvelope != 0);

  OTC_String theConversationId;
  theConversationId = theEnvelope->conversationId();

  OTCLIB_TRACER(MODULE) << "sendRequest =" << *theEnvelope << endl;
  OTCLIB_TRACER(MODULE) << "conversationId =" << theConversationId << endl;

  theEnvelope->queue(OTCLIB_STANDARD_JOB);

  return theConversationId;
}

/* ------------------------------------------------------------------------- */
void OTC_SVBroker::subscribeReports(
 char const* theServiceName,
 char const* theSubjectPattern
)
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_SVBroker::subscribeReports()");

  if (theServiceName == 0 || *theServiceName == EOS)
    return;

  OTC_String theKey;

  theKey += theServiceName;

  mutex_.lock();

  OTC_MutexReaper<OTC_NRMutex> xxxMutex;
  xxxMutex.grab(mutex_);

  if (!subscriptions_.contains(theKey,theSubjectPattern))
  {
    if (!subscriptions_.containsKey(theKey))
      registry()->subscribeServiceName(id(),theServiceName);

    subscriptions_.add(theKey,theSubjectPattern);

    OTC_Iterator<OTC_ServiceBinding*> thePublishers;
    thePublishers = publishers_.items(theKey);

    while (thePublishers.isValid())
    {
      // Send subscription request.

      OTCLIB_TRACER(MODULE) << "publisher = "
       << thePublishers.item()->serviceAddress() << endl;

      OTC_String theHeader;
      theHeader += "SubjectPattern: ";
      theHeader += theSubjectPattern;
      theHeader += EOL;

      OTC_Event* theMessage;
      theMessage = new OTCEV_Message("");
      OTCLIB_ASSERT_M(theMessage != 0);

      OTCEV_Envelope* theEnvelope;
      theEnvelope = new OTCEV_Envelope(theMessage,
       thePublishers.item()->serviceAddress(),
       serviceBinding_->serviceAddress(),
       "$S::SUBSCRIPTION-REQUESTED",theHeader);
      OTCLIB_ASSERT_M(theEnvelope != 0);

      theEnvelope->queue(OTCLIB_STANDARD_JOB);

      thePublishers.next();
    }
  }

  xxxMutex.release();

  mutex_.unlock();
}

/* ------------------------------------------------------------------------- */
void OTC_SVBroker::unsubscribeReports(
 char const* theServiceName,
 char const* theSubjectPattern
)
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_SVBroker::unsubscribeReports()");

  if (theServiceName == 0 || *theServiceName == EOS)
    return;

  OTC_String theKey;

  theKey += theServiceName;

  mutex_.lock();

  OTC_MutexReaper<OTC_NRMutex> xxxMutex;
  xxxMutex.grab(mutex_);

  if (subscriptions_.contains(theKey,theSubjectPattern))
  {
    subscriptions_.remove(theKey,theSubjectPattern);

    OTC_Iterator<OTC_ServiceBinding*> thePublishers;
    thePublishers = publishers_.items(theKey);

    while (thePublishers.isValid())
    {
      // Send the subscription cancellation.

      OTC_String theHeader;
      theHeader += "SubjectPattern: ";
      theHeader += theSubjectPattern;
      theHeader += EOL;

      OTC_Event* theMessage;
      theMessage = new OTCEV_Message("");
      OTCLIB_ASSERT_M(theMessage != 0);

      OTCEV_Envelope* theEnvelope;
      theEnvelope = new OTCEV_Envelope(theMessage,
       thePublishers.item()->serviceAddress(),
       serviceBinding_->serviceAddress(),
       "$S::SUBSCRIPTION-WITHDRAWN",theHeader);
      OTCLIB_ASSERT_M(theEnvelope != 0);

      theEnvelope->queue(OTCLIB_STANDARD_JOB);

      thePublishers.next();
    }

    if (!subscriptions_.containsKey(theKey))
      publishers_.removeKey(theKey,OTCLIB_UNSAFE);
  }

  if (!subscriptions_.containsKey(theKey))
    registry()->unsubscribeServiceName(id(),theServiceName);

  xxxMutex.release();

  mutex_.unlock();
}

/* ------------------------------------------------------------------------- */
void OTC_SVBroker::publishReport(
 char const* theSubjectName,
 OTC_ROPayload const& theContent,
 int theLifeTime
)
{
  OTCLIB_MARKBLOCK(MODULE,"OTC_SVBroker::publishReport()");

  OTC_String tmpContent;
  char const* tmpContentType;

  tmpContent = theContent.encode();
  tmpContentType = "text/xml";

  // Remove cache entry or update it.

  OTC_HIndexEntry<OTC_String>* theEntry;

  mutex_.lock();

  OTC_MutexReaper<OTC_NRMutex> xxxMutex;
  xxxMutex.grab(mutex_);

  if (theLifeTime != 0)
  {
    theEntry = subjects_.insert(theSubjectName);

    OTC_ReportCache* theCacheEntry;
    theCacheEntry = (OTC_ReportCache*)theEntry->item();

    if (theCacheEntry != 0)
      delete theCacheEntry;

    time_t thePostingTime;
    time_t theExpirationTime;

    thePostingTime = time(0);

    if (theLifeTime > 0)
      theExpirationTime = thePostingTime + theLifeTime;
    else
      theExpirationTime = (time_t)-1;

    theCacheEntry = new OTC_ReportCache(tmpContent,
     tmpContentType,thePostingTime,theExpirationTime);
    OTCLIB_ASSERT_M(theCacheEntry != 0);

    theEntry->setItem((void*)theCacheEntry);
  }
  else
  {
    theEntry = subjects_.entry(theSubjectName);

    if (theEntry != 0)
    {
      OTC_ReportCache* theCacheEntry;
      theCacheEntry = (OTC_ReportCache*)theEntry->item();

      delete theCacheEntry;

      theEntry->removeEntry();
    }
  }

  // Publish to subscribers with matching subject pattern.

  OTCEV_Message* theMessage;
  theMessage = new OTCEV_Message(tmpContent,tmpContentType);
  OTCLIB_ASSERT_M(theMessage != 0);

  OTC_PairIterator<OTC_String,OTC_ServiceBinding*> theSubscriptions;
  theSubscriptions = subscribers_.pairs();

  while (theSubscriptions.isValid())
  {
    if (matchSubject(theSubscriptions.key(),theSubjectName))
    {
      // Publish event.

      OTC_String theHeader;
      theHeader += "SubjectName: ";
      theHeader += theSubjectName;
      theHeader += EOL;

      OTCEV_Envelope* theEnvelope;
      theEnvelope = new OTCEV_Envelope(theMessage->clone(),
       theSubscriptions.item()->serviceAddress(),
       serviceBinding_->serviceAddress(),
       "$S::REPORT",theHeader);
      OTCLIB_ASSERT_M(theEnvelope != 0);

      theEnvelope->queue(OTCLIB_STANDARD_JOB);
    }

    theSubscriptions.next();
  }

  theMessage->destroy();

  xxxMutex.release();

  mutex_.unlock();
}

/* ------------------------------------------------------------------------- */
void OTC_SVBroker::sendReport(
 OTC_ServiceBinding* theServiceAgent,
 char const* theSubjectName,
 OTC_ROPayload const& theContent
)
{
  if (theServiceAgent == 0)
    return;

  OTC_String tmpContent;
  char const* tmpContentType;

  tmpContent = theContent.encode();
  tmpContentType = "text/xml";

  OTC_String theHeader;
  theHeader += "SubjectName: ";
  theHeader += theSubjectName;
  theHeader += EOL;

  OTCEV_Message* theMessage;
  theMessage = new OTCEV_Message(tmpContent,tmpContentType);
  OTCLIB_ASSERT_M(theMessage != 0);

  OTCEV_Envelope* theEnvelope;
  theEnvelope = new OTCEV_Envelope(theMessage,
   theServiceAgent->serviceAddress(),
   serviceBinding_->serviceAddress(),
   "$S::REPORT",theHeader);
  OTCLIB_ASSERT_M(theEnvelope != 0);

  theEnvelope->queue(OTCLIB_STANDARD_JOB);
}

/* ------------------------------------------------------------------------- */
void OTC_SVBroker::expireReports(char const* theSubjectPattern, u_int theAge)
{
  time_t theCurrentTime;
  theCurrentTime = time(0);

  mutex_.lock();

  OTC_MutexReaper<OTC_NRMutex> xxxMutex;
  xxxMutex.grab(mutex_);

  OTC_HIndexEntry<OTC_String>* theEntry;
  theEntry = subjects_.firstEntry();

  while (theEntry != 0)
  {
    OTC_ReportCache* theCacheEntry;
    theCacheEntry = (OTC_ReportCache*)theEntry->item();

    if (matchSubject(theSubjectPattern,theEntry->key()))
    {
      if (theCurrentTime > time_t(theCacheEntry->postingTime()+theAge))
      {
	delete theCacheEntry;

	OTC_HIndexEntry<OTC_String>* theOldEntry;
	theOldEntry = theEntry;

	theEntry = theEntry->nextEntry();

	theOldEntry->removeEntry();
      }
      else
      {
	theEntry = theEntry->nextEntry();
      }
    }
    else
    {
      theEntry = theEntry->nextEntry();
    }
  }

  xxxMutex.release();

  mutex_.unlock();
}

/* ------------------------------------------------------------------------- */
void OTC_SVBroker::expireReports(u_int theAge)
{
  time_t theCurrentTime;
  theCurrentTime = time(0);

  mutex_.lock();

  OTC_MutexReaper<OTC_NRMutex> xxxMutex;
  xxxMutex.grab(mutex_);

  OTC_HIndexEntry<OTC_String>* theEntry;
  theEntry = subjects_.firstEntry();

  while (theEntry != 0)
  {
    OTC_ReportCache* theCacheEntry;
    theCacheEntry = (OTC_ReportCache*)theEntry->item();

    if (theCurrentTime > time_t(theCacheEntry->postingTime()+theAge))
    {
      delete theCacheEntry;

      OTC_HIndexEntry<OTC_String>* theOldEntry;
      theOldEntry = theEntry;

      theEntry = theEntry->nextEntry();

      theOldEntry->removeEntry();
    }
    else
    {
      theEntry = theEntry->nextEntry();
    }
  }

  xxxMutex.release();

  mutex_.unlock();
}

/* ------------------------------------------------------------------------- */
void OTC_SVBroker::purgeReports()
{
  time_t theCurrentTime;
  theCurrentTime = time(0);

  mutex_.lock();

  OTC_MutexReaper<OTC_NRMutex> xxxMutex;
  xxxMutex.grab(mutex_);

  OTC_HIndexEntry<OTC_String>* theEntry;
  theEntry = subjects_.firstEntry();

  while (theEntry != 0)
  {
    OTC_ReportCache* theCacheEntry;
    theCacheEntry = (OTC_ReportCache*)theEntry->item();

    if (theCacheEntry->expirationTime() != -1 &&
     theCurrentTime > theCacheEntry->expirationTime())
    {
      delete theCacheEntry;

      OTC_HIndexEntry<OTC_String>* theOldEntry;
      theOldEntry = theEntry;

      theEntry = theEntry->nextEntry();

      theOldEntry->removeEntry();
    }
    else
    {
      theEntry = theEntry->nextEntry();
    }
  }

  xxxMutex.release();

  mutex_.unlock();
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_SVBroker::matchSubject(
 const char* theSubjectPattern,
 const char* theSubjectName
) const
{
  char c;

  while (1)
  {
    switch (c = *theSubjectPattern++)
    {
      case EOS:
      {
	// Must be end of string.

	return *theSubjectName == EOS;

	break;
      }

      case '?':
      {
	// Can't be end of string.

	if (*theSubjectName == EOS)
	  return OTCLIB_FALSE;

	++theSubjectName;

	break;
      }

      case '*':
      {
	// Use character following '*'.

	c = *theSubjectPattern;

	// Skip further instances of '*'.

	while (c == '*')
	  c = *++theSubjectPattern;

	// At end.

	if (c == EOS)
	  return OTCLIB_TRUE;

	// Use recursion for all other cases.

	while (*theSubjectName != EOS)
	{
	  if (OTC_SVBroker::matchSubject(theSubjectPattern,theSubjectName))
	    return OTCLIB_TRUE;

	  ++theSubjectName;
	}

	return OTCLIB_FALSE;

	break;
      }

      default:
      {
	if (c != *theSubjectName)
	  return OTCLIB_FALSE;

	theSubjectName++;

	break;
      }
    }
  }

  return OTCLIB_FALSE;
}

/* ------------------------------------------------------------------------- */
OTC_Boolean OTC_SVBroker::builtinRequest_(OTCEV_Request* theRequest)
{
  return OTCLIB_FALSE;

#if 0
  OTC_String theMethodName;
  theMethodName = theRequest->methodName();

  if (theMethodName.length() == 0)
    return OTCLIB_FALSE;

  if (theMethodName[0] != '_')
    return OTCLIB_FALSE;

  if (theMethodName == "_listReports")
  {
    OTC_SVPayload theResult;
    theResult <<= OTC_ROPayload::nullArray();

    mutex_.lock();

    OTC_MutexReaper<OTC_NRMutex> xxxMutex;
    xxxMutex.grab(mutex_);

    OTC_HIndexEntry<OTC_String>* theEntry;
    theEntry = subjects_.firstEntry();

    while (theEntry != 0)
    {
      OTC_ReportCache* theCacheEntry;
      theCacheEntry = (OTC_ReportCache*)theEntry->item();

      OTC_RWPayload theDetails;
      theDetails = theResult[-1];

      theDetails[-1] <<= theEntry->key();

      OTC_Time thePostingTime(
       theCacheEntry->postingTime(),OTCLIB_GMTIME);
      theDetails[-1] <<= thePostingTime;

      if (theCacheEntry->expirationTime() != -1)
      {
	OTC_Time theExpirationTime(
	 theCacheEntry->expirationTime(),OTCLIB_GMTIME);
	theDetails[-1] <<= theExpirationTime;
      }
      else
	theDetails[-1] <<= OTC_ROPayload::nullValue();

      theEntry = theEntry->nextEntry();
    }

    xxxMutex.release();

    mutex_.unlock();

    theRequest->sendResponse(theResult);
  }
  else
    return OTCLIB_FALSE;

  theRequest->destroy();

  return OTCLIB_TRUE;
#endif
}

/* ------------------------------------------------------------------------- */
